#!/usr/bin/env python3
# automome.py
# automome - the Ottish AUTOMEME clone
# Copyright (C) 2014  Penguin Development
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import copy
import getopt
import itertools
import os
import os.path
import random
import re
import sys

skipword = lambda x: ' '.join(x.split()[1:]) # Skips the first word

def skipuntilvowel(s):
    """Raptorish method to return the given string from the first vowel
    onwards. Returns the whole string if it has no vowels."""
    mat = re.compile(r'[AEIOU]').search(s)
    return s[mat.start():] if mat else s

# Lambda that turns a verb of form [x]ING into [x]ER:
erify = lambda x: (x[:-4] + 'IER') if x.endswith('YING') else (x[:-3] + 'ER')

class Replaceable(object):
    """Class for which a randomly picked word is substituted."""
    scanned = False
    scandir = '.'
    classes = {}

    def __init__(self, cls, default=None, prob=1.0):
        """Constructor. `cls' is the word class that this Replaceable should
        belong to; it should match the file name of a .dict.txt file in
        Replaceable.scandir. `cls' may also be a tuple or list of multiple
        word classes. `prob' is the probability of a replacement being chosen;
        (1-prob) is the probaility of `default' being used instead. If
        `default' evaluates to False, `prob' has no effect."""
        self.cls = (cls,) if isinstance(cls, str) else cls
        self.default = (default,) if isinstance(default, str) else default
        self.prob = prob
        self.scanclasses()
        if self.cls is None:
            return
        for c in self.cls:
            if c not in self.classes:
                raise KeyError(
                        'Mustard: word class `{}\' not found!'.format(c))

    @classmethod
    def scanclasses(cls, force = False):
        """Scans the directory Replaceable.scandir for files named *.dict.txt.
        The files' base names (i.e. the part before .dict.txt) are used as
        names for word classes; the files contain one word per line,
        whitespace is stripped off the start and end of each line, and empty
        lines (after stripping) are ignored.

        This method aborts immediately if it has been called previously,
        unless `force' evaluates to True."""
        if cls.scanned and not force:
            return

        for fn in os.listdir(cls.scandir):
            ffn = os.path.join(cls.scandir, fn)
            if fn.endswith('.dict.txt') and os.path.isfile(ffn):
                f = open(ffn, 'r', encoding='UTF-8')
                # the encoding in the line above is to prevent CGI mustard
                data = [x.strip() for x in f.read().splitlines()]
                f.close()
                data = [d for d in data if d and not d.startswith('#')]
                if len(data):
                    cls.classes[fn[:-9]] = data

        # Multiform nouns; index specification:
        # 0: singular w/ article
        # 1: plural
        cls.classes['noun-multiform'] = [[y.strip() for y in x.split('|')] for
                x in cls.classes['noun-multiform']]

        # Intransitive verbs; index specification:
        # 0: infinitive
        # 1: continuous
        # 2: present
        # 3: past
        # 4: perfect
        cls.classes['verb-intr-multiform'] = [[y.strip() for y in
            x.split('|')] for x in cls.classes['verb-intr-multiform']]

        # Transitive verbs; index specification:
        # 0: infinitive
        # 1: continuous
        # 2: present
        # 3: past
        # 4: perfect
        cls.classes['verb-tr-multiform'] = [[y.strip() for y in x.split('|')]
                for x in cls.classes['verb-tr-multiform']]

        # Noun classes
        cls.classes['noun-singular-article'] += \
            list(zip(*cls.classes['noun-multiform']))[0]

        cls.classes['noun-plural'] = \
            list(zip(*cls.classes['noun-multiform']))[1]

        cls.classes['noun-commodity'] = [skipword(x) for x in
            cls.classes['noun-singular-article']]

        cls.classes['noun-singular'] = cls.classes['noun-commodity'] + \
            cls.classes['noun-unique']

        cls.classes['noun-singular-1word'] = [x for x in
            cls.classes['noun-singular'] if len(x.split()) == 1]

        cls.classes['noun-unique-the'] = ['THE ' + x for x in
            cls.classes['noun-unique']]

        cls.classes['noun-singular-the'] = ['THE ' + x for x in
            cls.classes['noun-singular']]

        cls.classes['noun-plural-z'] = [x[:-1] + 'Z' for x in
            cls.classes['noun-plural'] if x.endswith('S')]

        # Intransitive verb classes
        cls.classes['verb-intr-inf'] = list(zip(
            *cls.classes['verb-intr-multiform']))[0]

        cls.classes['verb-intr-cont'] = list(zip(
            *cls.classes['verb-intr-multiform']))[1]

        cls.classes['verb-intr-present'] = list(zip(
            *cls.classes['verb-intr-multiform']))[2]

        cls.classes['verb-intr-past'] = list(zip(
            *cls.classes['verb-intr-multiform']))[3]

        cls.classes['verb-intr-perfect'] = list(zip(
            *cls.classes['verb-intr-multiform']))[4]

        cls.classes['verb-intr-inf-1word'] = [x for x in
            cls.classes['verb-intr-inf'] if len(x.split()) == 1]

        cls.classes['verb-intr-cont-in'] = [x[:-1] for x in
                cls.classes['verb-intr-cont'] if x.endswith('ING')]

        # Transitive verb classes
        cls.classes['verb-tr-inf'] = list(zip(
            *cls.classes['verb-tr-multiform']))[0]

        cls.classes['verb-tr-cont'] = list(zip(
            *cls.classes['verb-tr-multiform']))[1]

        cls.classes['verb-tr-present'] = list(zip(
            *cls.classes['verb-tr-multiform']))[2]

        cls.classes['verb-tr-past'] = list(zip(
            *cls.classes['verb-tr-multiform']))[3]

        cls.classes['verb-tr-perfect'] = list(zip(
            *cls.classes['verb-tr-multiform']))[4]

        cls.classes['verb-tr-inf-1word'] = [x for x in
            cls.classes['verb-tr-inf'] if len(x.split()) == 1]

        # -ING verbs
        cls.classes['verb-intr-ing-multiform'] = [x for x in
            cls.classes['verb-intr-multiform'] if x[1].endswith('ING')]
        cls.classes['verb-tr-ing-multiform'] = [x for x in
            cls.classes['verb-tr-multiform'] if x[1].endswith('ING')]
        cls.classes['verb-ing-multiform'] = \
            cls.classes['verb-intr-ing-multiform'] + \
            cls.classes['verb-tr-ing-multiform']

        # Adjective class
        cls.classes['adjective'] = [skipword(x) for x in
            cls.classes['adjective-article']]

        cls.scanned = True

    def __str__(self):
        """Returns a random word from the word class(es) self.cls that will be
        used in lieu of this Replaceable."""
        repl = ''
        if self.default and random.random() > self.prob:
            repl = random.choice(self.default)
        else:
            repl = random.choice(tuple(itertools.chain.from_iterable(
                [self.classes[c] for c in self.cls])))

        return repl

class MultiWord(object):
    """Class that holds different forms of the same word (e.g. singular and
    plural forms of nouns)."""

    def __init__(self, cls, idx = 0):
        """Constructor. `cls' is the multiform word class to use. `idx' is the
        default index (i.e. the default form) of the word that will be used."""
        if not cls.endswith('-multiform'):
            raise KeyError('Mustard: word class `{}\''
                ' is not a valid multiform class.'.format(cls))

        if not Replaceable.scanned:
            Replaceable.scanclasses()

        if cls not in Replaceable.classes:
            raise KeyError('Mustard: word class `{}\' not found!'.format(c))

        self.cls = cls
        self.index = idx
        self.word = None

    def __getitem__(self, idx):
        """Method that returns the word form corresponding to index `idx', or
        `self.index' if `idx' is None. If a word has not been picked, this
        method does so first; the word then remains fixed for the life-time of
        the MultiWord object."""

        if self.word is None:
            self.word = random.choice(Replaceable.classes[self.cls])

        return self.word[self.index if idx is None else idx]

class IndexMap(object):
    """Class that applies a function to the replacement, given the argument
    index of that replacement. E.g. let R = ('foo', IndexMap(0, f)) be the
    replacement tuple, then this will get expanded to ('foo', f('foo'))."""
    def __init__(self, idx, f = lambda x: x):
        self.index = idx
        self.func = f

    def apply(self, data):
        return self.func(data)

class Meme(object):
    """Meme class."""
    def __init__(self, template, *args):
        self.template = template
        self.args = args

    def __str__(self):
        """Returns a fully-filled version of the template."""
        args = [str(x) if not isinstance(x, (MultiWord, IndexMap)) else
                (copy.copy(x) if isinstance(x, MultiWord) else x) for x in
                self.args]
        for i in range(len(args)): # First loop: replace IndexMaps
            if isinstance(args[i], IndexMap):
                args[i] = args[i].apply(args[args[i].index]) # ARGCEPTION
        for i in range(len(args)): # Second loop: replace MultiWords with
                                   # their default form
            if isinstance(args[i], MultiWord):
                args[i] = args[i][None]

        try:
            return self.template.format(*args)
        except:
            print(">>> In template " + self.template, file=sys.stderr)
            raise


# Praparatory block to set the scan dir and seed the PRNG if automome.py is
# called directly.
if __name__ == '__main__':
    p = os.getenv('AUTOMOME_PATH')
    if (p):
        Replaceable.scandir = p
    elif not Replaceable.scandir:
        Replaceable.scandir = os.path.abspath(os.path.dirname(sys.argv[0]))

    random.seed()

R = lambda *x: Replaceable(x) # Replace with a word from the given classes
C = lambda *x: Replaceable(None, x, 0) # Replace with any of the chosen words
                                       # only
Rd = lambda d, p, *x: Replaceable(x, d, p) # Replace with a word from the
                                           # given classes, with default(s)

# For the origins of these memes, see mrob27's list at
#   http://mrob.com/time/automome/ott-snowclones.txt
memes = [
    Meme('404 - {} NOT FOUND',
        R('noun-singular')), # Non-AUTOMEME
    Meme('ALL YOUR {} ARE BELONG TO {}',
        R('noun-singular'),
        R('noun-singular', 'noun-plural', 'noun-mass', 'character')),
    Meme('ALSO, I CAN {} YOU WITH MY {}.',
        R('verb-tr-inf'),
        R('noun-singular')), # Non-AUTOMEME
    Meme('AND WHY DO YOU WORRY ABOUT {}? CONSIDER THE {} OF {}, '
            'HOW THEY {}: THEY NEITHER {} NOR {}.',
        R('noun-plural'),
        R('noun-plural'),
        R('place'),
        R('verb-tr-inf'),
        R('verb-tr-inf'),
        R('verb-tr-inf')), # Credit: mrob27
    Meme('ATTACK ITS WEAK POINT FOR {} {}',
        R('adjective'),
        R('noun-singular', 'noun-plural', 'noun-mass')),
    Meme('A WILD {} {}',
        R('noun-singular'),
        C('APPEARS', 'APPEARED')), # Credit: mrob27
    Meme('BECAUSE YOU ARE _{} {}_',
        Rd('HEADCRAB', 0.33, 'adjective'),
        R('noun-singular', 'noun-mass')),
    Meme('BETTER {} MY OWN {}',
        R('verb-tr-inf'),
        R('noun-singular', 'noun-mass')),
    Meme('BLESS YOU, {}. BL{}.',
        R('noun-plural', 'noun-mass', 'character'),
        IndexMap(0, skipuntilvowel)),
    Meme('BRACE {}, {} IS COMING',
        C('YOURSELF', 'YOURSELVES'),
        R('noun-singular', 'noun-mass')), # Credit: mrob27
    Meme('CH*RPING {}, HOW DO THEY WORK?',
        R('noun-plural')),
    Meme('CONSIDER THE {} OF {}',
        R('noun-plural', 'noun-mass'),
        R('place')), # Credit: mrob27
    Meme('COULD {} _BE_ ANY MORE {}?',
        R('noun-plural', 'noun-mass', 'character'),
        R('adjective')), # Non-AUTOMEME
    Meme('CROUCHING {}, HIDDEN {}',
        R('noun-singular'),
        R('noun-singular')), # Non-AUTOMEME
    Meme('DAMNIT JIM, I\'M {}, NOT {}',
        R('noun-singular-article'),
        R('noun-singular-article')),
    Meme('DIAL {} FOR {}',
        IndexMap(1, lambda x: x[0]),
        R('noun-singular', 'noun-plural', 'noun-mass', 'character')),
            # Non-AUTOMEME
    Meme('DISREGARD THAT, I {} {}',
        R('verb-tr-inf'),
        R('noun-plural', 'noun-mass')),
    Meme('DR. {}, OR HOW I LEARNED TO STOP {} AND LOVE THE {}',
        R('noun-singular', 'noun-mass', 'character'),
        Rd('WORRYING', 0.1, 'verb-intr-cont'),
        R('noun-singular')),
    Meme('EXCUSE ME {}, DO YOU HAVE A MOMENT TO TALK ABOUT {}?',
        C('SIR', 'MA\'AM'),
        R('character', 'noun-plural', 'noun-unique-the', 'noun-mass')),
            # Non-AUTOMEME
    Meme('GREAT LORD RANDALL IT\'S {} GET IN THE {}',
        R('noun-singular-article'),
        Rd('CAR', 0.33, 'noun-singular', 'noun-mass')),
    Meme('HELLO. MY NAME IS {}. YOU {} MY {}. PREPARE TO {}.',
        R('character'),
        R('verb-tr-past'),
        R('noun-singular', 'noun-plural', 'noun-mass'),
        R('verb-intr-inf')),
    Meme('HELLO? YES, THIS IS {}',
        R('noun-singular', 'noun-mass', 'character')),
    Meme('HOW DO I {} {}?',
        R('verb-tr-past'),
        R('noun-singular', 'noun-plural', 'noun-mass', 'character')),
    Meme('HOW IS {} FORMED? HOW {} GET {}?',
        R('noun-singular', 'noun-mass'),
        R('noun-singular', 'noun-mass', 'character'),
        Rd('PRAGNENT', 0.33, 'adjective')),
    Meme('I ACCIDENTALLY {}, IS THIS {}?',
        R('noun-singular-article'),
        C('BAD', 'BAD', 'DANGEROUS', 'ENDISH')),
    Meme('I AM {}, HEAR ME {}',
        R('character'),
        Rd('ROAR', 0.5, 'verb-intr-inf')),
    Meme('I CH*RPING LOVE {}',
        R('noun-plural', 'noun-mass')), # Non-AUTOMEME
    Meme('I DON\'T ALWAYS {}, BUT WHEN I DO, I {} {}',
        R('verb-intr-inf'),
        Rd('PREFER', 0.8, 'verb-tr-inf'),
        R('noun-plural', 'noun-mass')),
    Meme('I DON\'T ALWAYS {} {}, BUT WHEN I DO, I {} {}',
        R('verb-tr-inf'),
        R('noun-plural', 'noun-mass'),
        Rd('PREFER', 0.8, 'verb-tr-inf'),
        R('noun-plural', 'noun-mass')),
    Meme('I FIND YOUR LACK OF {} DISTURBING',
        R('noun-plural', 'noun-mass')), # Non-AUTOMEME
    Meme('I, FOR ONE, WELCOME OUR NEW {} OVERLORDS',
        R('noun-singular')),
    Meme('IF {} IS WRONG, I DON\'T WANT TO BE RIGHT',
        R('verb-tr-cont', 'verb-intr-cont', 'verb-tr-cont', 'verb-intr-cont',
            'character', 'noun-mass', 'noun-singular-the')), # Credit: mrob27
    Meme('I HAS {}. YOU CAN HAS MY {}.',
        R('noun-singular-article'),
        IndexMap(0, skipword)),
    Meme('I HAVE REPORTED YOU TO {} FOR {} {} AS IT IS A CRIME',
        R('character'),
        Rd('STEALING', 0.25, 'verb-tr-cont'),
        R('noun-plural', 'noun-mass')),
    Meme('I LIKE MY {} LIKE I LIKE MY {}: COVERED IN {}',
        Rd(('WOMEN', 'MEN'), 0.5, 'noun-plural', 'noun-mass'),
        R('noun-plural', 'noun-mass'),
        R('noun-plural', 'noun-mass')),
    Meme('I MADE YOU {} BUT I {} IT',
        R('noun-singular-article'),
        R('verb-tr-past')),
    Meme('IMMA CHARGIN MAH {}',
        R('noun-singular', 'noun-plural', 'noun-mass')),
    Meme('I\'M IN UR {}, {} UR {}',
        R('noun-singular', 'noun-mass'),
        R('verb-tr-cont'),
        R('noun-plural', 'noun-mass')),
    Meme('IN SOVIET RUSSIA, {} {} _YOU_',
        R('noun-singular', 'noun-mass', 'character'),
        R('verb-tr-present')),
    Meme('INSTANT {}, JUST ADD {}',
        R('noun-singular', 'noun-plural', 'noun-mass'),
        R('noun-plural', 'noun-mass')), # Non-AUTOMEME
    Meme('I SEE WHAT YOU {} THERE',
        R('verb-tr-past')),
    Meme('IS IT CAN BE {} TIEM NAO PLEES?',
        R('noun-singular', 'noun-plural-z')),
    Meme('I THINK {} IS A PRETTY {} {}. '
            'XEH {} {} AND DOESN\'T AFRAID OF ANYTHING',
        R('character'),
        Rd(('TREEISH', 'MOLPISH'), 0.2, 'adjective'),
        Rd('XUY', 0.1, 'noun-singular'),
        R('verb-tr-present'),
        R('noun-plural', 'noun-mass', 'character')),
    Meme('IT\'S OVER. {} IS FINISHED.',
        R('noun-singular', 'noun-mass', 'character')),
    Meme('IT\'S {} ALL THE WAY DOWN',
        R('noun-plural', 'noun-mass')), # Credit: mrob27
    Meme('IT\'S {}. IT\'S {}. IT\'S _SUPER{}_!',
        R('noun-singular-article'),
        R('noun-singular-article'),
        R('noun-singular-1word')), # Non-AUTOMEME
    Meme('I USED TO BE {} LIKE YOU, THEN I TOOK {} IN THE KNEE',
        R('noun-singular-article'),
        R('noun-singular-article')), # Non-AUTOMEME
    Meme('I WEAR {} NOW. {} ARE {}.',
        MultiWord('noun-multiform', 0),
        IndexMap(0, lambda x: x[1]),
        C('COOL', 'COOL', 'TREEISH', 'MOLPISH')), # Non-AUTOMEME
    Meme('I WILL NOT {} THIS {} -- IT IS {}',
        R('verb-tr-inf'),
        R('noun-singular', 'noun-mass'),
        R('noun-singular-article', 'noun-mass', 'adjective', 'character')),
    Meme('I {} _ALL_ THE {}',
        R('verb-tr-past'),
        R('noun-plural', 'noun-mass')),
    Meme('I {}, THEREFORE I {}',
        Rd('THINK', 0.5, 'verb-intr-inf'),
        R('verb-intr-inf')),
    Meme('KEEP CALM AND {} ON',
        Rd('MOLPY', 0.9, 'verb-intr-inf-1word')), # Credit: mrob27
    Meme('MAXIMUM {} YIELDS MAXIMUM {}',
        R('noun-plural', 'noun-mass'),
        R('noun-plural', 'noun-mass')),
    Meme('MY {} BRINGS ALL THE {} TO {}',
        R('noun-singular'),
        R('noun-plural'),
        Rd('THE YARD', 0.5, 'place')), # Non-AUTOMEME
    Meme('MY {} IS FULL OF {}',
        R('noun-singular', 'noun-mass'),
        R('noun-plural', 'noun-mass')),
    Meme('MY {} JUST {} ME; M*STARD WAS SO {}',
        R('noun-singular', 'noun-mass'),
        R('verb-tr-past'),
        Rd('CASH', 0.5, 'adjective')),
    Meme('MY {}, LET ME SHOW YOU IT',
        R('noun-singular', 'noun-mass')),
    Meme('NEVER SEND TO KNOW FOR WHOM THE {0} {1}; IT {1} FOR THEE',
        R('noun-singular'),
        R('verb-intr-present', 'verb-tr-present')), # Credit: mrob27
    Meme('NI NI {} {} {}',
        C('NI', 'NI', 'NI NI', 'NI NI NI'),
        Rd('CHUPACABRA', 0.8, 'noun-singular', 'noun-mass'),
        Rd('PING PONG BALL', 0.8, 'noun-singular', 'noun-mass')), # Meta-meme
    Meme('NOOOO THEY BE {} MY {}',
        R('verb-tr-cont'),
        R('noun-singular', 'noun-plural', 'noun-mass')),
    Meme('NOT SURE IF {} OR JUST {}',
        R('verb-intr-cont'),
        R('verb-intr-cont', 'adjective')),
    Meme('NOW I AM BECOME {}, THE {} OF {}',
        R('character', 'noun-singular', 'noun-mass'),
        IndexMap(-1, lambda x: erify(x[1])),
        R('noun-plural', 'noun-mass'),
        MultiWord('verb-tr-ing-multiform', 1)), # Non-AUTOMEME
    Meme('NOW IF YOU\'LL EXCUSE ME, I HAVE {} TO {}',
        R('noun-singular-article', 'noun-plural', 'noun-mass'),
        R('verb-tr-inf')), # Non-AUTOMEME
    Meme('NO, {}, YOU ARE THE {}',
        R('noun-singular', 'noun-mass', 'character'),
        R('noun-singular', 'noun-mass')),
    Meme('OH GLR HOW DID {} GET HERE I AM NOT GOOD WITH {}',
        Rd('THIS', 0.5, 'noun-plural', 'noun-mass', 'character'),
        R('noun-plural', 'noun-mass')),
    Meme('OH HAI, I {} YOUR {}',
        R('verb-tr-past'),
        R('noun-singular', 'noun-plural', 'noun-mass')),
    Meme('ONE DOES NOT SIMPLY {} INTO {}',
        R('verb-intr-inf'),
        Rd('MORDOR', 0.3, 'place')), # Credit: mrob27
    Meme('ONE {} DOES NOT {} MAKE',
        R('noun-commodity'),
        R('noun-singular-article')), # Non-AUTOMEME
    Meme('PH\'NGLUI MGLW\'NAFH {} {} WGAH\'NAGL FHTAGN',
        R('character'),
        IndexMap(-1, lambda x: x[4:] if x.startswith('THE ') else x),
        R('place')), # Non-AUTOMEME
    Meme('PLEASE READ: A PERSONAL APPEAL FROM {} FOUNDER {}.',
        R('noun-singular'),
        R('character')), # Credit: mrob27
    Meme('PROTIP: TO {} THE {}, {} IT UNTIL IT {}',
        Rd('DEFEAT', 0.75, 'verb-tr-inf'),
        R('noun-singular', 'noun-mass'),
        R('verb-tr-inf'),
        R('verb-intr-present')),
    Meme('PUSH {}, RECEIVE {}',
        C('BUTAN', 'BUTTON'),
        R('noun-singular', 'noun-plural', 'noun-mass')),
    Meme('REMAIN {} AND {} {}',
        Rd('CALM', 0.8, 'adjective'),
        R('verb-tr-inf'),
        R('noun-plural', 'noun-mass')), # Credit: mrob27
    Meme('SCREAMING {}: {} DOESN\'T LIKE IT',
        R('noun-singular', 'character'),
        R('character')),
    Meme('SO I HERD U {} {}',
        Rd('LIEK', 0.25, 'verb-tr-inf'),
        R('noun-plural', 'noun-mass')),
    Meme('SOMETIMES YOU {0} THE {1}, SOMETIMES THE {1} {2} YOU',
        MultiWord('verb-tr-multiform', 0),
        R('noun-singular'),
        IndexMap(0, lambda x: x[2])), # Non-AUTOMEME
    Meme('SPEAK {} AND CARRY {} {}',
        Rd('SOFTLY', 0.2, 'adverb'),
        Rd('A SEAISH', 0.67, 'adjective-article'),
        R('noun-singular')), # None-AUTOMEME
    Meme('SUDDENLY, {}. THOUSANDS OF THEM!',
        R('noun-plural')), # Non-AUTOMEME
    Meme('THANKS, {}. TH{}.',
        R('noun-plural', 'noun-mass', 'character'),
        IndexMap(0, skipuntilvowel)),
    Meme('THANK YOU {}, BUT OUR {} IS IN ANOTHER {}',
        R('character'),
        R('noun-singular', 'noun-mass'),
        R('noun-singular')),
    Meme('THAT\'S NOT {0}; _THIS_ IS {0}',
        R('noun-singular-article')),
    Meme('THAT\'S ONE RIVERISH {0} FOR {1}, ONE SEAISH {0} FOR {2}KIND',
        R('noun-singular'),
        R('noun-singular-article'),
        IndexMap(1, skipword)), # Non-AUTOMEME
    Meme('THE FIRST RULE OF {0} IS, YOU DON\'T TALK ABOUT {0}',
        R('noun-mass')), # Credit: mrob27
    Meme('THE FIRST RULE OF {0} CLUB IS, YOU DON\'T TALK ABOUT {0} CLUB',
        R('noun-singular')), # Credit: mrob27
    Meme('THERE IS {} IN MY {}, YOUR ARGUMENT IS INVALID',
        R('noun-singular-article'),
        Rd('BEARD', 0.67, 'noun-singular', 'noun-mass')),
    Meme('THESE AREN\'T THE {} YOU\'RE {} FOR',
        R('noun-plural'),
        R('verb-intr-cont', 'verb-tr-cont')), # Credit: mrob27
    Meme('THEY\'RE TAKING THE {} TO ISENGARD!',
        R('noun-plural')), # Credit: mrob27
    Meme('THEY SEE ME {}, THEY {}',
        R('verb-intr-cont-in'),
        Rd('HATIN', 0.67, 'verb-intr-cont-in')),
    Meme('THE {} IS A LIE',
        R('noun-singular')),
    Meme('THE {} IS TOO CH*RPING {}',
        R('noun-singular'),
        Rd('HIGH', 0.7, 'adjective')), # Credit: mrob27
    Meme('"THIS IS NOT ACCEPTABLE!" I SCREAMED AS {} {}',
        R('character'),
        R('verb-intr-past')),
    Meme('THIS IS YOUR {} ON {}',
        Rd('BRAIN', 0.25, 'noun-singular'),
        R('noun-plural', 'noun-mass')),
    Meme('THIS IS {}. I CAN TELL BY THE {}, '
            'AND FROM HAVING SEEN A LOT OF {} IN MY DIP',
        Rd(('PHOTOSHOPPED', 'OTTIFIED'), 0.67, 'adjective',
            'verb-tr-perfect'),
        R('noun-plural', 'noun-mass'),
        R('noun-plural', 'noun-mass')),
    Meme('TO {0} OR NOT TO {0}, THAT IS THE QUESTION',
            R('verb-tr-inf', 'verb-intr-inf')), # Credit: mrob27
    Meme('VERY IMPORTANT FOR {} TO {} XIM {}',
        R('character'),
        Rd('DO', 0.75, 'verb-tr-inf'),
        Rd('JOB', 0.5, 'noun-singular', 'noun-mass')),
    Meme('WE\'RE GONNA NEED A MORE {} {}',
        Rd('SEAISH', 0.3, 'adjective'),
        R('noun-singular')), # Credit: mrob27
    Meme('WE\'VE SECRETLY REPLACED {}\'S {} WITH {}. '
            'LET\'S SEE IF XE NOTICES...',
        R('character'),
        R('noun-singular', 'noun-mass'),
        R('noun-singular-article')),
    Meme('WHAT _ARE_ {}? WE JUST DON\'T KNOW.',
        R('noun-plural')),
    Meme('WHATEVER {} YOUR {}',
        R('verb-tr-present'),
        Rd('RAFTCASTLE', 0.2, 'noun-singular', 'noun-plural', 'noun-mass')),
    Meme('WHAT HAPPENS IN {0}, STAYS IN {0}',
        R('place')),
    Meme('WHAT THE CH*RP? YOU ARE {}. AND {}',
        R('noun-singular', 'noun-mass', 'character'),
        R('adjective')),
    Meme('WITH {} LIKE THESE, WHO NEEDS {}?',
        R('noun-plural'),
        R('noun-plural', 'noun-mass')), # Non-AUTOMEME
    Meme('WOW - SO {} - MANY {} - SUCH {} - VERY {} - MUCH {} - {}',
        R('noun-singular'),
        R('adjective'),
        R('verb-intr-cont', 'verb-tr-cont'),
        R('noun-singular'),
        R('adjective'),
        C('SEA IS BIG', 'RIVER IS SMALL', 'HAIL RANDALL')), # Credit: mrob27
    Meme('Y\'ALL GOT ANY MORE OF THEM {}?',
        R('noun-plural')), # Non-AUTOMEME
    Meme('YOU CAN TAKE {0} OUT OF {1}, BUT YOU CAN\'T TAKE {1} OUT OF {0}',
        R('character', 'noun-singular-the'),
        R('place')), # Non-AUTOMEME
    Meme('YOU HAVE {} MOST OF A RIVERISH {}. '
            'PLEASE PICK YOUR {} WITH GREATER CARE.',
        R('verb-tr-past'),
        R('noun-commodity'),
        R('noun-plural', 'noun-mass')),
    Meme('YOU WOULDN\'T {} {}',
        Rd('DOWNLOAD', 0.5, 'verb-tr-inf'),
        R('noun-singular-article')),
    Meme('YO WAWLPY, I HEARD YOU LIKE {0} SO WE PUT SOME {0} IN YOUR {0} '
            'SO YOU CAN {1} WHILE YOU {1}',
        R('noun-plural', 'noun-mass'),
        R('verb-intr-inf')),
    Meme('YO {}, I\'M REALLY HAPPY FOR YOU, IMMA LET YOU FINISH, '
            'BUT {} HAD ONE OF THE {} {} OF ALL TIME',
        R('character'),
        R('character'),
        C('MOST TREEISH', 'MOLPIEST', 'MOST WOWTERFUL'),
        R('noun-plural')),
    Meme('_{} 2: ELECTRIC BAGALOO_',
        R('noun-singular', 'noun-mass', 'noun-plural', 'verb-tr-cont',
        'verb-tr-inf')), # Credit: mrob27
    Meme('_{0} HARD 2: {0} HARDER_',
        R('verb-tr-inf-1word', 'verb-intr-inf-1word')), # Non-AUTOMEME
    Meme('{} AND {} AND {}, OH MY!',
        R('noun-plural'),
        R('noun-plural'),
        R('noun-plural', 'noun-mass')), # Credit: mrob27
    Meme('{} ARE THE NEW {}',
        R('noun-plural'),
        R('noun-plural')), # Credit: mrob27
    Meme('{0}MEOWLPY IS {1}',
        IndexMap(1, skipword),
        R('noun-singular-article')),
    Meme('{} DON\'T KNOW ABOUT MY {}',
        R('noun-plural'),
        R('noun-singular', 'noun-plural', 'noun-mass')),
    Meme('{} GRAB MY {}',
        R('character'),
        R('noun-singular', 'noun-plural')), # Non-AUTOMEME
    Meme('{}? IN _MY_ {}?',
        R('noun-plural', 'noun-mass'),
        R('noun-singular', 'noun-mass')),
    Meme('{} IS {} TOO',
        R('noun-singular-article'),
        R('adjective')),
    Meme('{} IS SERIOUS {}',
        R('noun-singular', 'character'),
        R('noun-singular')),
    Meme('{} IS THE NEW {}',
        R('noun-singular', 'noun-mass'),
        R('noun-singular', 'noun-mass')), # Credit: mrob27
    Meme('{} IS THE {} OF THE {}',
        R('noun-singular', 'character'),
        R('noun-singular', 'noun-mass'),
        R('noun-plural', 'noun-mass')),
    Meme('{0}MAN, {0}MAN, DOES WHATEVER {1} CAN',
        IndexMap(1, skipword),
        R('noun-singular-article')), # Non-AUTOMEME
    Meme('{}\'S CLOSED DUE TO {}',
        Rd('POOL', 0.5, 'noun-singular', 'noun-mass'),
        R('noun-plural', 'noun-mass', 'character')),
    Meme('{} OR GTCO',
        R('noun-plural', 'noun-mass')),
    Meme('{0} THE {1}, {0} THE WORLD',
        Rd('SAVE', 0.2, 'verb-tr-inf'),
        R('noun-singular', 'noun-plural', 'noun-mass')), # Non-AUTOMEME
    Meme('{} THE {}... WITH YOUR {}',
        R('verb-tr-inf'),
        R('noun-plural', 'noun-mass'),
        R('noun-singular', 'noun-plural', 'noun-mass')),
    Meme('{}, WE HAVE {}SIGN THE LIKES OF WHICH EVEN GLR HAS NEVER SEEN!',
        R('character'),
        R('noun-singular-1word')), # Non-AUTOMEME
    Meme('{} {} ARE _SUPERIOR_',
        R('adjective'),
        R('noun-plural')),
    Meme('{} {} GONNA {}',
        MultiWord('verb-ing-multiform', 1),
        IndexMap(0, lambda x: (erify(x[1]) + 'S')),
        IndexMap(0, lambda x: x[0])), # Credit: mrob27
    Meme('{0} {0} {0} {0} {0} {0} {0} {0} {0} {0} {0} {0} {1} {1}',
        Rd('MOLPY', 0.8, 'noun-singular-1word'),
        Rd('GRAPEVINE', 0.8, 'noun-singular-1word'))
            # Meta-meme; credit: mrob27
]

memes_unmolpish = [
    Meme('BITCHES DON\'T KNOW ABOUT MY {}',
        R('noun-singular', 'noun-plural', 'noun-mass')),
    Meme('HELLO. MY NAME IS {}. YOU KILLED MY {}. PREPARE TO {}.',
        R('character'),
        R('noun-singular', 'noun-plural', 'noun-mass'),
        Rd('DIE', 0.1, 'verb-intr-inf')),
    Meme('I\'M {0}! I\'M {0}! SUCK MY DIIIIICK! I\'M {0}!',
        R('noun-singular-article', 'character')),
    Meme('PROTIP: TO {} THE {}, {} IT UNTIL IT DIES',
        Rd('DEFEAT', 0.75, 'verb-tr-inf'),
        R('noun-singular', 'noun-mass'),
        R('verb-tr-inf')),
    Meme('{} HAET {}',
        Rd('LOLI', 0.5, 'character'),
        R('noun-singular', 'noun-plural', 'noun-mass', 'character')),
    Meme('{} IS PIG DISGUSTING',
        R('noun-singular', 'character')),
    Meme('{} IS THE {} THAT IS CH*RPING {}',
        R('noun-singular', 'noun-mass', 'character'),
        Rd('BACONSEMENCOFFEEBABYCANCER', 0.1, 'noun-singular', 'noun-mass'),
        R('noun-plural', 'noun-mass', 'character')),
    Meme('{} KILLS {}',
        R('noun-singular', 'character'),
        R('noun-plural', 'character')),
    Meme('{0} YOUR CH*RPING {1}. {0} THEM.',
        R('verb-tr-inf'),
        R('noun-plural'))
]

def print_help(f=sys.stdout):
    print("""Usage: automome.py [option] [num]
[option] may be one of the following:
    -h, --help              show this message and exit
    -V, --version           show version and exit
    -u, --unmolpish         include unmolpish memes

[num] is the number of memes to generate; if it is missing or invalid, it is
assumed to be 1.""", file=f)

def print_version():
    print("""automome version 2.1, the Ottish AUTOMEME clone
Copyright (C) 2014 Penguin Development
Inspired by AUTOMEME (http://automeme.net/)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.""")

if __name__ == '__main__':
    n = 1
    unmolpish = False
    opts, args = getopt.gnu_getopt(sys.argv[1:], 'hVu', ('help', 'version',
        'unmolpish'))

    for o, a in opts:
        if o in ('-h', '--help'):
            print_help()
            sys.exit(0)
        elif o in ('-V', '--version'):
            print_version()
            sys.exit(0)
        elif o in ('-u', '--unmolpish'):
            unmolpish = True

    if len(args) > 0:
        n = int(args[0])

    if n < 1:
        n = 1

    if unmolpish:
        memes += memes_unmolpish

    for i in range(n):
        print(str(random.choice(memes)))
